package project

import (
	"database/sql"

	"github.com/go-gorp/gorp"
	"github.com/ovh/cds/engine/api/application"
	"github.com/ovh/cds/engine/api/environment"
	"github.com/ovh/cds/engine/api/group"
	"github.com/ovh/cds/engine/api/integration"
	"github.com/ovh/cds/engine/api/pipeline"
	"github.com/ovh/cds/engine/api/workflow"
	"github.com/ovh/cds/sdk"
)

// LoadOptionFunc is used as options to loadProject functions
type LoadOptionFunc func(gorp.SqlExecutor, *sdk.Project) error

// LoadOptions provides all options on project loads functions
var LoadOptions = struct {
	Default                                 LoadOptionFunc
	WithIcon                                LoadOptionFunc
	WithApplications                        LoadOptionFunc
	WithApplicationNames                    LoadOptionFunc
	WithVariables                           LoadOptionFunc
	WithVariablesWithClearPassword          LoadOptionFunc
	WithPipelines                           LoadOptionFunc
	WithPipelineNames                       LoadOptionFunc
	WithEnvironments                        LoadOptionFunc
	WithEnvironmentNames                    LoadOptionFunc
	WithGroups                              LoadOptionFunc
	WithPermission                          LoadOptionFunc
	WithApplicationVariables                LoadOptionFunc
	WithApplicationKeys                     LoadOptionFunc
	WithApplicationWithDeploymentStrategies LoadOptionFunc
	WithKeys                                LoadOptionFunc
	WithWorkflows                           LoadOptionFunc
	WithWorkflowNames                       LoadOptionFunc
	WithClearKeys                           LoadOptionFunc
	WithIntegrations                        LoadOptionFunc
	WithClearIntegrations                   LoadOptionFunc
	WithFavorites                           func(uID string) LoadOptionFunc
	WithLabels                              LoadOptionFunc
}{
	Default:                                 loadDefault,
	WithIcon:                                loadIcon,
	WithPipelines:                           loadPipelines,
	WithPipelineNames:                       loadPipelineNames,
	WithEnvironments:                        loadEnvironments,
	WithEnvironmentNames:                    loadEnvironmentNames,
	WithGroups:                              loadGroups,
	WithApplications:                        loadApplications,
	WithApplicationNames:                    loadApplicationNames,
	WithVariables:                           loadVariables,
	WithVariablesWithClearPassword:          loadVariablesWithClearPassword,
	WithApplicationVariables:                loadApplicationVariables,
	WithApplicationKeys:                     loadApplicationKeys,
	WithKeys:                                loadKeys,
	WithWorkflows:                           loadWorkflows,
	WithWorkflowNames:                       loadWorkflowNames,
	WithClearKeys:                           loadClearKeys,
	WithIntegrations:                        loadIntegrations,
	WithClearIntegrations:                   loadClearIntegrations,
	WithFavorites:                           loadFavorites,
	WithApplicationWithDeploymentStrategies: loadApplicationWithDeploymentStrategies,
	WithLabels:                              loadLabels,
}

func loadDefault(db gorp.SqlExecutor, proj *sdk.Project) error {
	if err := loadVariables(db, proj); err != nil {
		return sdk.WithStack(err)
	}
	if err := loadApplications(db, proj); err != nil {
		return sdk.WithStack(err)
	}
	return nil
}

func loadApplications(db gorp.SqlExecutor, proj *sdk.Project) error {
	if err := loadApplicationsWithOpts(db, proj); err != nil {
		return sdk.WithStack(err)
	}
	return nil
}

func loadApplicationNames(db gorp.SqlExecutor, proj *sdk.Project) error {
	apps, err := application.LoadAllNames(db, proj.ID)
	if err != nil {
		return sdk.WithStack(err)
	}
	proj.ApplicationNames = apps
	return nil
}

func loadApplicationWithDeploymentStrategies(db gorp.SqlExecutor, proj *sdk.Project) error {
	if proj.Applications == nil {
		if err := loadApplications(db, proj); err != nil {
			return sdk.WithStack(err)
		}
	}
	for i := range proj.Applications {
		a := &proj.Applications[i]
		if err := (*application.LoadOptions.WithDeploymentStrategies)(db, a); err != nil {
			return sdk.WithStack(err)
		}
	}
	return nil
}

func loadVariables(db gorp.SqlExecutor, proj *sdk.Project) error {
	vars, err := LoadAllVariables(db, proj.ID)
	if err != nil {
		return err
	}
	proj.Variables = vars
	return nil
}

func loadVariablesWithClearPassword(db gorp.SqlExecutor, proj *sdk.Project) error {
	vars, err := LoadAllVariablesWithDecrytion(db, proj.ID)
	if err != nil {
		return err
	}
	proj.Variables = vars
	return nil

}

func loadApplicationVariables(db gorp.SqlExecutor, proj *sdk.Project) error {
	if proj.Applications == nil {
		if err := loadApplications(db, proj); err != nil {
			return sdk.WithStack(err)
		}
	}

	for _, a := range proj.Applications {
		if err := (*application.LoadOptions.WithVariables)(db, &a); err != nil {
			return sdk.WithStack(err)
		}
	}

	return nil
}

func loadApplicationKeys(db gorp.SqlExecutor, proj *sdk.Project) error {
	if proj.Applications == nil {
		if err := loadApplications(db, proj); err != nil {
			return sdk.WithStack(err)
		}
	}

	for _, a := range proj.Applications {
		if err := (*application.LoadOptions.WithKeys)(db, &a); err != nil {
			return sdk.WithStack(err)
		}
	}

	return nil
}

func loadKeys(db gorp.SqlExecutor, proj *sdk.Project) error {
	keys, err := LoadAllKeys(db, proj.ID)
	if err != nil {
		return err
	}
	proj.Keys = keys
	return nil
}

func loadClearKeys(db gorp.SqlExecutor, proj *sdk.Project) error {
	keys, err := LoadAllKeysWithPrivateContent(db, proj.ID)
	if err != nil {
		return err
	}
	proj.Keys = keys
	return nil
}

func loadIntegrations(db gorp.SqlExecutor, proj *sdk.Project) error {
	pf, err := integration.LoadIntegrationsByProjectID(db, proj.ID)
	if err != nil {
		return sdk.WrapError(err, "cannot load integrations")
	}
	proj.Integrations = pf
	return nil
}

func loadClearIntegrations(db gorp.SqlExecutor, proj *sdk.Project) error {
	pf, err := integration.LoadIntegrationsByProjectIDWithClearPassword(db, proj.ID)
	if err != nil {
		return sdk.WrapError(err, "cannot load integrations")
	}
	proj.Integrations = pf
	return nil
}

func loadWorkflows(db gorp.SqlExecutor, proj *sdk.Project) error {
	workflows, err := workflow.LoadAll(db, proj.Key)
	if err != nil {
		return sdk.WithStack(err)
	}
	proj.Workflows = workflows
	return nil
}

func loadWorkflowNames(db gorp.SqlExecutor, proj *sdk.Project) error {
	wfs, err := workflow.LoadAllNames(db, proj.ID)
	if err != nil {
		return sdk.WithStack(err)
	}
	proj.WorkflowNames = wfs
	return nil
}

func loadApplicationsWithOpts(db gorp.SqlExecutor, proj *sdk.Project, opts ...application.LoadOptionFunc) error {
	apps, err := application.LoadAll(db, proj.Key, opts...)
	if err != nil && sdk.Cause(err) != sql.ErrNoRows && !sdk.ErrorIs(err, sdk.ErrNotFound) {
		return sdk.WithStack(err)
	}
	proj.Applications = apps
	return nil
}

func loadIcon(db gorp.SqlExecutor, proj *sdk.Project) error {
	icon, err := db.SelectStr("SELECT icon FROM project WHERE id = $1", proj.ID)
	if err != nil {
		return sdk.WithStack(err)
	}
	proj.Icon = icon
	return nil
}

func loadPipelines(db gorp.SqlExecutor, proj *sdk.Project) error {
	pipelines, err := pipeline.LoadPipelines(db, proj.ID, false)
	if err != nil && sdk.Cause(err) != sql.ErrNoRows && !sdk.ErrorIs(err, sdk.ErrPipelineNotFound) && !sdk.ErrorIs(err, sdk.ErrPipelineNotAttached) {
		return sdk.WithStack(err)
	}
	proj.Pipelines = append(proj.Pipelines, pipelines...)
	return nil
}

func loadPipelineNames(db gorp.SqlExecutor, proj *sdk.Project) error {
	pips, err := pipeline.LoadAllNames(db, proj.ID)
	if err != nil {
		return sdk.WithStack(err)
	}
	proj.PipelineNames = pips
	return nil
}

func loadEnvironments(db gorp.SqlExecutor, proj *sdk.Project) error {
	envs, err := environment.LoadEnvironments(db, proj.Key)
	if err != nil && sdk.Cause(err) != sql.ErrNoRows && !sdk.ErrorIs(err, sdk.ErrNoEnvironment) {
		return sdk.WithStack(err)
	}
	proj.Environments = envs
	return nil
}

func loadEnvironmentNames(db gorp.SqlExecutor, proj *sdk.Project) error {
	var err error
	var envs sdk.IDNames

	if envs, err = environment.LoadAllNames(db, proj.ID); err != nil {
		return sdk.WrapError(err, "cannot load environment names")
	}
	proj.EnvironmentNames = envs

	return nil
}

func loadGroups(db gorp.SqlExecutor, proj *sdk.Project) error {
	if err := group.LoadGroupsIntoProject(db, proj); err != nil && sdk.Cause(err) != sql.ErrNoRows {
		return err
	}
	return nil
}

func loadLabels(db gorp.SqlExecutor, proj *sdk.Project) error {
	labels, err := Labels(db, proj.ID)
	if err != nil {
		return sdk.WithStack(err)
	}
	proj.Labels = labels
	return nil
}

func loadFavorites(uID string) LoadOptionFunc {
	return func(db gorp.SqlExecutor, proj *sdk.Project) error {
		count, err := db.SelectInt("SELECT COUNT(1) FROM project_favorite WHERE project_id = $1 AND authentified_user_id = $2", proj.ID, uID)
		if err != nil {
			return sdk.WithStack(err)
		}
		proj.Favorite = count > 0
		return nil
	}
}
