package workflow

import (
	"context"
	"database/sql"
	"time"

	"github.com/go-gorp/gorp"
	"github.com/lib/pq"
	"github.com/rockbears/log"

	"github.com/ovh/cds/sdk"
)

// ExistsStepLog returns the size of step log if exists.
func ExistsStepLog(db gorp.SqlExecutor, id int64, order int64) (bool, int64, error) {
	query := `
    SELECT octet_length(value) as size
    FROM workflow_node_run_job_logs
    WHERE workflow_node_run_job_id = $1 AND step_order = $2
  `

	var size int64
	if err := db.QueryRow(query, id, order).Scan(&size); err != nil {
		if sdk.Cause(err) != sql.ErrNoRows {
			return false, 0, sdk.WithStack(err)
		}
		return false, 0, nil
	}

	return true, size, nil
}

//LoadStepLogs load logs (workflow_node_run_job_logs) for a job (workflow_node_run_job) for a specific step_order
func LoadStepLogs(db gorp.SqlExecutor, id int64, order int64) (*sdk.Log, error) {
	log.Debug(context.TODO(), "LoadStepLogs> workflow_node_run_job_id = %d", id)
	query := `
		SELECT id, workflow_node_run_job_id, workflow_node_run_id, start, last_modified, done, step_order, value
		FROM workflow_node_run_job_logs
		WHERE workflow_node_run_job_id = $1 AND step_order = $2`
	logs := &sdk.Log{}
	var s, m, d pq.NullTime
	if err := db.QueryRow(query, id, order).Scan(&logs.ID, &logs.JobID, &logs.NodeRunID, &s, &m, &d, &logs.StepOrder, &logs.Val); err != nil {
		if err == sql.ErrNoRows {
			return nil, nil
		}
		return nil, err
	}

	if s.Valid {
		logs.Start = &s.Time
	}
	if m.Valid {
		logs.LastModified = &m.Time
	}
	if d.Valid {
		logs.Done = &d.Time
	}
	return logs, nil
}

//LoadLogs load logs (workflow_node_run_job_logs) for a job (workflow_node_run_job)
func LoadLogs(db gorp.SqlExecutor, id int64) ([]sdk.Log, error) {
	query := `
		SELECT id, workflow_node_run_job_id, workflow_node_run_id, start, last_modified, done, step_order, value
		FROM workflow_node_run_job_logs
		WHERE workflow_node_run_job_id = $1
		ORDER BY id`
	rows, err := db.Query(query, id)
	if err != nil {
		return nil, err
	}
	defer rows.Close() // nolint
	var logs []sdk.Log
	for rows.Next() {
		l := &sdk.Log{}
		var s, m, d pq.NullTime

		if err := rows.Scan(&l.ID, &l.JobID, &l.NodeRunID, &s, &m, &d, &l.StepOrder, &l.Val); err != nil {
			return nil, err
		}

		if s.Valid {
			l.Start = &s.Time
		}
		if m.Valid {
			l.LastModified = &m.Time
		}
		if d.Valid {
			l.Done = &d.Time
		}

		logs = append(logs, *l)
	}
	return logs, nil
}

func insertLog(db gorp.SqlExecutor, logs *sdk.Log) error {
	now := time.Now()
	logs.Start = &now
	logs.LastModified = &now

	query := `
		INSERT INTO workflow_node_run_job_logs (workflow_node_run_job_id, workflow_node_run_id, start, last_modified, done, step_order, value)
		VALUES ($1, $2, $3, $4, $5, $6, $7)
		RETURNING ID `
	return sdk.WithStack(db.QueryRow(query, logs.JobID, logs.NodeRunID, logs.Start, logs.LastModified, logs.Done, logs.StepOrder, logs.Val).Scan(&logs.ID))
}

func updateLog(db gorp.SqlExecutor, logs *sdk.Log) error {
	now := time.Now()
	logs.LastModified = &now
	logs.Done = &now

	query := `
		UPDATE workflow_node_run_job_logs set
			workflow_node_run_id = $3,
			last_modified = $4,
			done = $5,
			value = value || $6
		WHERE workflow_node_run_job_id = $1 AND step_order = $2`

	if _, err := db.Exec(query, logs.JobID, logs.StepOrder, logs.NodeRunID, logs.LastModified, logs.Done, logs.Val); err != nil {
		return sdk.WithStack(err)
	}
	return nil
}
